#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts:
# Creation Date: 2021-03-12
# Language: Python3

################################### IMPORTS ####################################

# Standard library
# Your imports from the standard library go here


# External imports
# Your imports from other packages go here


# Internal imports
from compass.ast.expression import Expression  # Used for inheritance

################################### CLASSES ####################################


class Signal(Expression):
    """
    Class used to represent a Signal in the AST. A signal is like a
    non-persistent variable that our program will emit or react to.
    """

    def __init__(self, name: str):
        """
        Constructor of the Signal class.

        Arguments
        =========
         - name: The name used to identify this Signal.
        """
        # Storing our arguments.
        self.name = name

    # To simplify the codegen slightly, we gove a string representation to Signals.
    def __repr__(self) -> str:
        """
        String representation of the Signal.
        """
        return self.name


################################## FUNCTIONS ###################################

# Your functions go here

##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################



#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts:
# Creation Date: 2021-03-18
# Language: Python3

################################### IMPORTS ####################################

# Standard library
from __future__ import annotations  # Used for self reference
from typing import Optional  # Used for type hints


# External imports
# Your imports from other packages go here


# Internal imports
from compass.ast.statement import Statement  # Used for inheritance
from compass.ast.expression import Expression  # Used for type hints

################################### CLASSES ####################################


class IfStatement(Statement):
    """
    Class used to represent an if-else statement in the AST.
    """

    def __init__(
        self,
        expression: Expression,
        statement: Statement,
        else_statement: Optional[Statement] = None,
    ):
        """
        Constructor of the Await class.

        Arguments
        =========
         - expression: The Expression for the condition to execute the
            statement.
         - statement: The statement to execute if the condition is met.
        """
        # Storing the arguments.
        self.expression = expression
        self.statement = statement
        self.else_statement = else_statement

    def with_else(self, statement: Statement) -> IfStatement:
        """
        Adds the provided statement as an else statement to the current one and
        returns the updated statement with the two branches.

        Arguments
        =========
         - statement: The statement to use in the else branch.
        """
        # We return a new IfStatement to avoid side effects.
        return IfStatement(
            self.expression, self.statement, else_statement=statement
        )


################################## FUNCTIONS ###################################

# Your functions go here

##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################



#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts:
# Creation Date: 2021-03-30
# Language: Python3

################################### IMPORTS ####################################

# Standard library
from __future__ import annotations  # Used for self reference
from typing import List, Dict, Any  # Used for type hints


# External imports
# Your imports from other packages go here


# Internal imports
# Your imports within this package go here

################################### CLASSES ####################################


class BuilderConfiguration:
    """
    Class used to hold the configuration of the builder that was stored in the
    JSON file.
    """

    def __init__(
        self,
        entry_point: str,
        debug: bool,
        modules: Dict[str, List[str]],
        source_directory: str,
        build_directory: str,
    ):
        """
        Constructors of the BuilderConfiguration class.

        Arguments
        =========
         - entry_points: The module that will be used as an entry-point for your
            program.
         - debug: Whether the CLI debugging wrapper should be computed.
         - modules: A dictionary associating each source module with the
            instances it should be compiled to.
         - source_directory: Path to the directory where the compass files can
            be found.
         - build_directory: Path to the directory where the compiled files
            should be placed.
        """
        # Storing the arguments.
        self.entry_point = entry_point
        self.debug = debug
        self.modules = modules
        self.source_directory = source_directory
        self.build_directory = build_directory

    def to_makefile(self) -> str:
        """
        Compiles this BuilderConfiguration into a Makefile template that can be
        used for configuration.
        """
        source_code = ""
        # Adding a small header to the Makefile.
        source_code += "# Makefile compiled automatically by compass-build.\n\n"
        # Declaring generic aliases to make the built Makefile easier to edit.
        source_code += f"SRC_DIR = {self.source_directory}\n"
        source_code += f"BUILD_DIR = {self.build_directory}\n"
        # We need to at least give the "-c" flag to the compiler to produce
        # object files.
        source_code += f"COMPASS_FLAGS = -c $(CFLAGS)\n"
        source_code += build_source_aliases(self.modules)
        # Blank line for the style.
        source_code += "\n"
        source_code += build_destination_aliases(self.modules)
        # Blank line for the style.
        source_code += "\n"
        source_code += build_entry_point_aliases(
            self.modules, self.entry_point, self.debug
        )
        # Blank line for the style.
        source_code += "\n"
        # Line for separation.
        source_code += "#" * 80 + "\n\n"
        # Adding the code in the Makefile for the rules. We start with the PHONY
        # modifier.
        source_code += ".PHONY: all clean\n\n"
        # Adding the rules for the entry point first.
        source_code += build_entry_point_rule(self.entry_point, self.debug)
        # Adding the rules for all the object files of the module instances.
        source_code += build_module_rules(self.modules)
        # Adding a rule to build the build directory if it doesn't exist.
        source_code += "$(BUILD_DIR):\n" "\tmkdir -p $(BUILD_DIR)\n\n"
        # Adding the rules to clean the code.
        source_code += build_clean_rule()
        # Returning the built Makefile.
        return source_code

    @classmethod
    def from_json(cls, json_content: Any) -> BuilderConfiguration:
        """
        Deserializes the BuildConfiguration from the content of the provided
        Makefile. Will raise an AssertionError if deserialization fails.
        """
        # The root of the JSON file should be a dictionary.
        assert isinstance(json_content, dict)
        # The root of the JSON should have 2 or 3 keys.
        number_keys = len(json_content.keys())
        assert number_keys == 4 or number_keys == 5
        # Getting the entry point.
        assert "entry_point" in json_content
        # Getting the value for the entry_point.
        entry_point: Any = json_content["entry_point"]
        # The entry point should be a string.
        assert isinstance(entry_point, str)
        # Getting the module dictionary.
        assert "modules" in json_content
        modules: Any = json_content["modules"]
        # The modules value should be a dictionary.
        assert isinstance(modules, dict)
        # The keys for the modules should all be string.
        assert all(isinstance(key, str) for key in modules.keys())
        # The values for the modules should all be lists.
        assert all(isinstance(value, list) for value in modules.values())
        # The renamed modules in the value lists should all be strings.
        assert all(
            all(isinstance(renamed_module, str) for renamed_module in value)
            for value in modules.values()
        )
        # Getting the source directory.
        assert "source_directory" in json_content
        # Getting the value for the source_directory.
        source_directory: Any = json_content["source_directory"]
        # The source directory should be a string.
        assert isinstance(source_directory, str)
        # Getting the build directory.
        assert "build_directory" in json_content
        # Getting the value for the build_directory.
        build_directory: Any = json_content["build_directory"]
        # The build directory should be a string.
        assert isinstance(build_directory, str)
        # We try to get the debug value if there is one.
        if number_keys == 5:
            assert "debug" in json_content
            debug: Any = json_content["debug"]
            # debug should be either True or False.
            assert isinstance(debug, bool)
        else:
            # No debug by default.
            debug = False
        # We return an instance of the BuilderConfiguration class from the
        # fetched values.
        return cls(
            entry_point, debug, modules, source_directory, build_directory
        )


################################## FUNCTIONS ###################################


def build_source_aliases(modules: Dict[str, List[str]]) -> str:
    """
    Returns the Makefile snippet for the aliases of all the described source
    modules.
    """
    source_code = ""
    # We first fetch all the source modules.
    source_modules = modules.keys()
    for source_module in source_modules:
        # We create an alias for the source module. We assume that the file for
        # the provided module can be derived from the module name.
        source_code += (
            f"SRC_{source_module.upper()} = $(SRC_DIR)/{source_module}.cmps\n"
        )
    # We return the built source code.
    return source_code


def build_destination_aliases(modules: Dict[str, List[str]]) -> str:
    """
    Returns the Makefile snippet for the aliases of all the described module
    instances.
    """
    source_code = ""
    # We first fetch all the destination modules. We assume that no 2 modules
    # have the same name. We concatenate all the module instances in one big
    # list.
    instances = sum(modules.values(), [])
    # We also keep a list of all the instances we will need for the "all" rule.
    all_instances: List[str] = list()
    # We also need to keep track of the intermediary C files for the "clean"
    # rule.
    all_c_instances: List[str] = list()
    for instance in instances:
        # We create an alias for the provided instance and generate a name for
        # its object file. We also add an alias for its C code.
        source_code += (
            f"DEST_{instance.upper()} = $(BUILD_DIR)/{instance}.o\n"
            f"DEST_C_{instance.upper()} = $(BUILD_DIR)/{instance}.c\n"
        )
        # We add the alias for the object to the list of instances.
        all_instances.append(f"$(DEST_{instance.upper()})")
        # Same for the C file.
        all_c_instances.append(f"$(DEST_C_{instance.upper()})")
    # We build an alias to build all the object files at once.
    source_code += "ALL_DEST = " + " ".join(all_instances) + "\n"
    # Same for the intermediary C files.
    source_code += "ALL_DEST_C = " + " ".join(all_c_instances) + "\n"
    # We return the built snippet.
    return source_code


def build_entry_point_aliases(modules: Dict[str, List[str]], entry_point: str, debug: bool) -> str:
    """
    Returns the Makefile snippet to build the aliases used in the compilation of
    the entry point module.
    """
    source_code = ""

    # We start by finding which module the entry point comes from.
    source_module: Optional[str] = None
    for module, instances in modules.items():
        if entry_point in instances:
            # We have found the module from which the entry point originates,
            # we remember it and exit.
            source_module = module
            break
    # We raise an assertion error if the entry_point is never produced.
    assert source_module is not None, f"{entry_point} is never produced"

    # We add an alias to link the source file of the entry point.
    source_code += f"SRC_ENTRY_POINT = $(SRC_{source_module.upper()})\n"
    # We create an alias for the entry point header. Default header name has a
    # "_compass" at the end, this is to play nicely with the debug backend and
    # avoid name collision.
    source_code += f"HEAD_ENTRY_POINT = $(BUILD_DIR)/{entry_point}_compass.h\n"
    # If required, we add an alias for the CLI debuggin wrapper.
    if debug:
        source_code += (
            f"DEBUG_ENTRY_POINT = $(BUILD_DIR)/{entry_point}.debug.o\n"
            f"DEBUG_C_ENTRY_POINT = $(BUILD_DIR)/{entry_point}.debug.c\n"
            f"DEBUG_ELF_ENTRY_POINT = $(BUILD_DIR)/{entry_point}.debug.elf\n"
        )
    # We return the prepared source code.
    return source_code


def build_entry_point_rule(entry_point: str, debug: bool) -> str:
    """
    Returns the rule to build the entry point of the project.
    """
    source_code = ""
    # We start by making the "all" entry rule. We don't add an EOL because of
    # the debug flag. We also build the "build directory" if it doesn't yet
    # exist.
    source_code += "all: $(ALL_DEST) $(HEAD_ENTRY_POINT) $(BUILD_DIR)"
    if debug:
        # We also have to add the debug entry-point to the all rule.
        source_code += " $(DEBUG_ELF_ENTRY_POINT)"
    # Now that the debug has been handled, we may close the "all" rule.
    source_code += "\n\n"

    # We add the rule to build the header of the entry point.
    source_code += (
        "$(HEAD_ENTRY_POINT): $(SRC_ENTRY_POINT)\n"
        f"\tcompass -r {entry_point} -o $(HEAD_ENTRY_POINT) --lang header "
        "$(SRC_ENTRY_POINT)\n\n"
    )

    if debug:
        # If the debug flag is set, we also have to compile the debugging CLI.
        # We need two rules for this. The first one is to built the C source for
        # the CLI.
        source_code += (
            "$(DEBUG_C_ENTRY_POINT): $(SRC_ENTRY_POINT) $(BUILD_DIR)\n"
            f"\tcompass -r {entry_point} -o $(DEBUG_C_ENTRY_POINT) --lang "
            "debug $(SRC_ENTRY_POINT)\n\n"
        )
        # The second rule is to build the object file from the CLI C source. We
        # need to include the build directory since it is where the header file
        # can be found.
        source_code += (
            "$(DEBUG_ENTRY_POINT): $(DEBUG_C_ENTRY_POINT) $(BUILD_DIR)\n"
            "\t$(CC) $(COMPASS_FLAGS) -o $(DEBUG_ENTRY_POINT) -I $(BUILD_DIR) "
            "$(DEBUG_C_ENTRY_POINT)\n\n"
        )
        # Because the debug file contains the main for the code, we may compile
        # the ELF directly. We include a third rule for this purpose.
        source_code += (
            "$(DEBUG_ELF_ENTRY_POINT): $(DEBUG_ENTRY_POINT) $(ALL_DEST)\n"
            "\t$(CC) $(CFLAGS) -o $(DEBUG_ELF_ENTRY_POINT) "
            "$(DEBUG_ENTRY_POINT) $(ALL_DEST)\n\n"
        )

    # We return the prepared source code.
    return source_code


def build_module_rules(modules: Dict[str, List[str]]) -> str:
    """
    Adds a Makefile snippet with the rules to compile all the required modules.
    """
    source_code = ""
    # We add rules for the modules, one by one.
    for source_module, instances in modules.items():
        # Each module might need to be compiled to several instances.
        for instance in instances:
            # Alias to the C IR.
            c_alias = f"DEST_C_{instance.upper()}"
            # Alias to the compiled object file.
            object_alias = f"DEST_{instance.upper()}"
            # Alias to the compass source file.
            compass_alias = f"SRC_{source_module.upper()}"
            # We add a rule to compile the C code for the module from the compass
            # source code.
            source_code += (
                f"$({c_alias}): $({compass_alias}) $(BUILD_DIR)\n"
                f"\tcompass -r {instance} -o $({c_alias}) --lang C "
                f"$({compass_alias})\n\n"
            )
            # We add a second rule to make the object file from the C intermediate
            # representation.
            source_code += (
                f"$({object_alias}): $({c_alias}) $(BUILD_DIR)\n"
                f"\t$(CC) $(COMPASS_FLAGS) -o $({object_alias}) $({c_alias})\n\n"
            )
    # We return the prepared Makefile snippet.
    return source_code


def build_clean_rule() -> str:
    """
    Builds the Makefile snippet for the "clean" rule of the project.
    """
    # We return the simple snippet for the clean rule.
    return (
        "clean:\n"
        "\trm -f $(ALL_DEST) $(ALL_DEST_C) $(HEAD_ENTRY_POINT) "
        "$(DEBUG_ENTRY_POINT) $(DEBUG_C_ENTRY_POINT)\n\n"
    )


##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################



#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts:
# Creation Date: 2021-03-12
# Language: Python3

################################### IMPORTS ####################################

# Standard library
# Your imports from the standard library go here


# External imports
# Your imports from other packages go here


# Internal imports
from compass.ast.statement import Statement  # Used for inheritance
from compass.ast.expression import Expression  # Used for type hints

################################### CLASSES ####################################


class Each(Statement):
    """
    Class used to represent an Each loop in the AST.
    """

    def __init__(self, expression: Expression, statement: Statement):
        """
        Constructor of the Each class.

        Arguments
        =========
         - expression: The Expression to use as a reset for this loop.
         - statement: The Statement which constitutes the body of this loop.
        """
        # Storing the arguments.
        self.expression = expression
        self.statement = statement


################################## FUNCTIONS ###################################

# Your functions go here

##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################



#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts:
# Creation Date: 2021-03-12
# Language: Python3

################################### IMPORTS ####################################

# Standard library
# Your imports from the standard library go here


# External imports
# Your imports from other packages go here


# Internal imports
import compass.ast as ast  # Used for downcasts
from compass.codegen.compass.module import (
    codegen_module,
)  # Used for recursive codegen
from compass.codegen.compass.declaration import (
    codegen_declaration,
)  # Used for recursive codegen

################################### CLASSES ####################################

# Your classes go here

################################## FUNCTIONS ###################################


def codegen_tree(tree: ast.Tree) -> str:
    """
    Outputs the Compass source code for the provided AST.
    """
    source_code = ""

    # We start by building the code for all the extern declarations.
    for extern in tree.externs:
        # We build the source for the arguments of the extern submodule.
        code_arguments = ", ".join(
            [
                codegen_declaration(declaration)
                for declaration in extern.declarations
            ]
        )
        # We add the source code for the entire declaration of the extern
        # submodule.
        source_code += f"extern {extern.name}({code_arguments});\n"
    # If there are no extern modules, we just add a comment.
    if len(tree.externs) == 0:
        source_code += "# No extern submodules."

    # A blank line for the style.
    source_code += "\n"

    # We add the source code for the main module.
    source_code += codegen_module(tree.module)

    # We return the complete source code.
    return source_code


##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################



#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts:
# Creation Date: 2021-03-12
# Language: Python3

################################### IMPORTS ####################################

# Standard library
from typing import Optional  # Used for type hints


# External imports
# Your imports from other packages go here


# Internal imports
from compass.grammar import CompassLexer, CompassParser  # Used to build the AST
from compass.ast import Tree  # Used for type hints

# The codegen functions for the different targets.
from compass.codegen.compass import codegen_tree as codegen_compass
from compass.codegen.c import codegen_tree as codegen_c
from compass.codegen.header import codegen_tree as codegen_header
from compass.codegen.debug import codegen_tree as codegen_debug

################################### CLASSES ####################################

# Your classes go here

################################## FUNCTIONS ###################################


def compass_compile(
    source_code: str, target: str, rename: Optional[str]
) -> str:
    """
    Runs the compilation process for the provided source code.

    Arguments
    =========
     - source_code: The source code to compile.
     - target: The target to compile to. Valid targets are detailed in the
        documentation.
     - rename: The new name to use for the module in the compiled code. This is
        usefull to create several modules from the same source code.

    Returns
    =======
    The compiled source code.
    """
    # We first build the tokens from the source code.
    tokens = CompassLexer().tokenize(source_code)
    # We then build the AST from the tokens.
    ast = CompassParser().parse(tokens)

    # If the parsing step failed because of syntax errors, the AST will be None
    # here.
    if ast is None:
        raise ValueError("Source code could not be parsed")

    # We rename the compiled module if we were asked to.
    if rename is not None:
        ast.rename(rename)

    # The we turn the AST into source code using the right codegen based on the
    # provided target.
    if target == "compass":
        return codegen_compass(ast)
    elif target in ["C", "c"]:
        return codegen_c(ast)
    elif target in ["H", "h", "header"]:
        return codegen_header(ast)
    elif target == "debug":
        return codegen_debug(ast)
    else:
        raise ValueError(f"Unknown target {target}")


##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################



#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts:
# Creation Date: 2021-03-12
# Language: Python3

################################### IMPORTS ####################################

# Standard library
import logging  # Used to print error message, but keep parsing the source


# External imports
from sly import Lexer  # Lexer implementation


# Internal imports
from compass.grammar import compass_tokens  # Used for language tokens

################################### CLASSES ####################################


class CompassLexer(Lexer):
    """
    Lexer for the Compass language.
    """

    # The tokens used to build the source code.
    tokens = compass_tokens.tokens
    # Ignoring whitespace and tab.
    ignore = " \t"
    # Ignoring everything in a line after a '#' (for comments).
    ignore_comment = r"\#.*"
    # Some symbols are used literaly.
    literals = compass_tokens.literals

    # Matching the assignation operator.
    ASSIGN = r"<-"
    # Matching non-literal the binary operators.
    TESTEQ = r"=="
    AND = r"&&"
    OR = r"\|\|"

    # Matching a number, only integers are allowed in compass.
    NUMBER = r"\d+"
    # Matching any string, which is first assumed to be a name.
    NAME = r"[a-zA-Z_][a-zA-Z0-9_]*"

    # Remapping the language keywords when they are found within the string.
    NAME["module"] = MODULE
    NAME["endmodule"] = ENDMODULE
    NAME["submodule"] = SUBMODULE
    NAME["input"] = INPUT
    NAME["output"] = OUTPUT
    NAME["each"] = EACH
    NAME["seq"] = SEQ
    NAME["par"] = PAR
    NAME["await"] = AWAIT
    NAME["emit"] = EMIT
    NAME["local"] = LOCAL
    NAME["if"] = IF
    NAME["elif"] = ELIF
    NAME["else"] = ELSE
    NAME["endif"] = ENDIF
    NAME["extern"] = EXTERN

    # We keep track of newlines to count them.
    @_(r"\n+")
    def newline(self, t):
        self.lineno += len(t.value)

    # Used to print error messages when lexer error is encountered.
    def error(self, t):
        # Logging an error message.
        logging.error(f"Unrecognized string {t.value}")
        self.index += len(t.value)


################################## FUNCTIONS ###################################

# Your functions go here

##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################



#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts:
# Creation Date: 2021-03-13
# Language: Python3

################################### IMPORTS ####################################

# Standard library
from typing import List, Tuple  # Used for type hints


# External imports
# Your imports from other packages go here


# Internal imports
import compass.ast as ast  # Used for downcasts
from compass.codegen.c.declaration import (
    codegen_declaration,
)  # Used for recursive codegen
from compass.codegen.c.statement import (
    codegen_statement,
)  # Used for recursive codegen
from compass.codegen.utils import (
    InnerContext,
    OuterContext,
)  # Used for type hints.

################################### CLASSES ####################################

# Your classes go here

################################## FUNCTIONS ###################################


def codegen_module(module: ast.Module) -> Tuple[str, List[str], List[str]]:
    """
    Outputs the C code corresponding to the provided Module.

    Returns
    =======
    A tuple with two values:
     - The source code for the inner module.
     - The states that have to be statically allocated for the module.
     - The local variables that have to be statically allocated for the module.
    """
    source_code = ""
    # Adding the name of the module.
    source_code += f"void {module.name}("
    # Building the code for each declaration.
    code_declarations = [
        codegen_declaration(declaration) for declaration in module.declarations
    ]
    # Joining the codes for the declarations.
    source_code += ", ".join(code_declarations)
    # Opening the brackets for the body of the code.
    source_code += ") {\n"
    # Adding the (indented) code for the body of the module.
    ic = codegen_statement(module.statement, OuterContext(indent=1))
    source_code += ic.source_code
    # Closing the brackets.
    source_code += "}\n"
    # Returing the prepared source code.
    return source_code, ic.owned_states, ic.owned_locals


##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################



#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts:
# Creation Date: 2021-03-17
# Language: Python3

################################### IMPORTS ####################################

# Standard library
# Your imports from the standard library go here


# External imports
# Your imports from other packages go here


# Internal imports
# Your imports within this package go here

################################### CLASSES ####################################


class OuterContext:
    """
    Class used to provide several interesting values during the recursive
    compilation of compass statements. The variable will often be called "oc" in
    the python code.
    """

    def __init__(self, indent: int = 0):
        """
        Constructor of the OuterContext class.

        Arguments
        =========
         - indent: The current level of indentation.
        """
        # Storing the arguments.
        self.indent = indent
        # Creating the indentation string.
        self.indent_str = indent * "\t"


################################## FUNCTIONS ###################################

# Your functions go here

##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################



#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts: 
# Creation Date: 2021-03-13
# Language: Python3

################################### IMPORTS ####################################

# Standard library 
# Your imports from the standard library go here 


# External imports 
# Your imports from other packages go here 


# Internal imports 
import compass.ast as ast  # Used for downcasts
from compass.codegen.header.module import (
    codegen_module,
)  # Used for recursive codegen


################################### CLASSES ####################################

# Your classes go here 

################################## FUNCTIONS ###################################

def codegen_tree(tree: ast.Tree) -> str:
    """
    Outputs the C source code for the provided AST.
    """
    source_code = ""
    # We have to add a include header guard. The string used for the header
    # guard is built from the name of the module.
    header_guard = f"{tree.module.name}_COMPASS_INCLUDED".upper()
    # Adding the guard to the header.
    source_code += f"#ifndef {header_guard}\n"
    source_code += f"#define {header_guard}\n"
    # Newline for the style.
    source_code += "\n"
    # For now an AST only contains one module.
    source_code += codegen_module(tree.module)
    # Newline for the style.
    source_code += "\n"
    # Closing the include guard.
    source_code += "#endif\n"
    # Returning the built source code.
    return source_code

##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################



#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts:
# Creation Date: 2021-03-13
# Language: Python3

################################### IMPORTS ####################################

# Standard library
from typing import List, Tuple  # Used for type hints


# External imports
# Your imports from other packages go here


# Internal imports
import compass.ast as ast  # Used for downcasts
from compass.codegen.c.expression import (
    codegen_expression,
)  # Used for recursive codegen
from compass.codegen.utils import (
    InnerContext,
    OuterContext,
)  # Used for type hints

################################### CLASSES ####################################


class SSAGenerator:
    """
    Static class used to build a singleton that outputs new unique names.
    """

    # The counter used to build the unique names.
    COUNTER = 0

    @classmethod
    def new_name(cls, base_name: str) -> str:
        """
        Returns a new unique name based on the provided one.
        """
        # We build the new name from the counter.
        new_name_string = f"{base_name}_{SSAGenerator.COUNTER}"
        # We increment the counter for the next call.
        SSAGenerator.COUNTER += 1
        # We return the built string.
        return new_name_string


################################## FUNCTIONS ###################################


def codegen_statement(
    statement: ast.Statement, oc: OuterContext
) -> InnerContext:
    """
    Outputs the Compass code corresponding to the provided Statement.

    Arguments
    =========
     - statement: The statement to build the code for.
     - oc: The OuterContext we have to take into account when building this statement.

    Returns
    =======
    The InnerContext produced by this code (which includes the generated code).

    Note
    ====
    All the real states are always contained in the owned states. Immediate
    states are boolena values telling whether a statement is done or not and can
    be rvalues. Owned states are used for initialization and resets.
    """
    # Several cases appear, one for each kind of statements.
    if isinstance(statement, ast.Each):
        return codegen_each(statement, oc)
    elif isinstance(statement, ast.Seq):
        return codegen_seq(statement, oc)
    elif isinstance(statement, ast.Par):
        return codegen_par(statement, oc)
    elif isinstance(statement, ast.AwaitStatement):
        return codegen_await(statement, oc)
    elif isinstance(statement, ast.EmitStatement):
        return codegen_emit(statement, oc)
    elif isinstance(statement, ast.LocalStatement):
        return codegen_local(statement, oc)
    elif isinstance(statement, ast.IfStatement):
        return codegen_if(statement, oc)
    elif isinstance(statement, ast.Submodule):
        return codegen_submodule(statement, oc)
    else:
        raise ValueError(f"Unknown statement {statement}")


def codegen_each(statement: ast.Each, oc: OuterContext) -> InnerContext:
    """
    Specialized variant of codegen_statement for for-each statements.
    """
    # The Inner Context we will return. Note that an each loop never ends.
    ic = InnerContext(immediate_state="0")
    # We build the code for the inner body of the for-each loop.
    inner_ic = codegen_statement(statement.statement, oc)
    # We build the code for the reset expression.
    inner_expression = codegen_expression(statement.expression)
    # for-each statements don't use the inner_immediate states in any particular
    # way.
    ic.inherit(inner_ic)
    # The each loops don't have any state, but we have to build a bit of code
    # for the reset.
    ic.source_code += oc.indent_str + f"if ({inner_expression}) {{\n"
    # We have to reset all the states owned by the body of the loop.
    for state in ic.owned_states:
        ic.source_code += oc.indent_str + "\t" + f"{state} = 0;\n"
    # Closing the reset statement and beginning the statement of the body.
    ic.source_code += oc.indent_str + "}\n"
    # Adding the code for the inner body.
    ic.source_code += inner_ic.source_code
    # Returning the source code, the immediate states and the owned states.
    return ic


def codegen_seq(statement: ast.Seq, oc: OuterContext) -> InnerContext:
    """
    Specialized variant of codegen_statement for seq statements.
    """
    # The Inner Context we will return.
    ic = InnerContext()
    # We need a state to keep track of the sequential execution.
    seq_state = SSAGenerator.new_name("seq")
    # We add our new state to the list of states we need to keep track of.
    ic.owned_states.append(seq_state)
    # We build some code for each of the inner sequential statements. We first
    # have to create the context for those future statement.
    inner_oc = OuterContext(indent=oc.indent + 1)
    for index, inner_statement in enumerate(statement.statements):
        # We build the code for the inner body of the seq statement.
        inner_ic = codegen_statement(inner_statement, inner_oc)
        # We inherit the recursive values from the InnerContext.
        ic.inherit(inner_ic)
        # We add the code for the case. Note that we always use ifs (and never
        # else) to fall to the next case whenever possible.
        ic.source_code += oc.indent_str + f"if ({seq_state} == {index}) {{\n"
        # We add the source code for the statement executed sequentially.
        ic.source_code += inner_ic.source_code
        # EDGE CASE
        # If we have reached the last state of the sequential execution, we
        # don't jump.
        if index + 1 == len(statement.statements):
            ic.source_code += (
                oc.indent_str + "\t" + "/* LAST SEQUENTIAL STATE */\n"
            )
        else:
            # We move to the next state if all the immediate states of the inner
            # statement are true, or if there are no immediate states.
            ic.source_code += (
                oc.indent_str
                + "\t"
                + f"{seq_state} += {inner_ic.immediate_state};\n"
            )
        # Closing the if statement.
        ic.source_code += oc.indent_str + "}\n"
    # We return the expected source code. The single immediate state of a seq
    # statement is whether it reached its last state.
    ic.immediate_state = (f"({seq_state} == {len(statement.statements) - 1}",)
    return ic


def codegen_par(statement: ast.Par, oc: OuterContext) -> InnerContext:
    """
    Specialized variant of codegen_statement for par statements.
    """
    # The Inner Context we will return.
    ic = InnerContext()
    # We also keep track of all the immediate states of our sub statements to
    # build our own immediate state at the end.
    immediate_states: List[str] = list()
    # Par statements are mostly invisible in the produced code, they look like
    # normal C execution.
    # We build some code for each of the inner parallel statements.
    for index, inner_statement in enumerate(statement.statements):
        # We build the code for the inner body of the par statement. We don't
        # need to indent the code.
        inner_ic = codegen_statement(inner_statement, oc)
        # We inherit the recursive values from the InnerContext.
        ic.inherit(inner_ic)
        # We also keep track of the immediate state.
        immediate_states.append(inner_ic.immediate_state)
        # We add the source code for the statement executed sequentially.
        ic.source_code += inner_ic.source_code
    # We return the expected source code. The single immediate state of a par
    # statement is the intersection of all the immediate states of its children
    # (i.e. the parallel execution ends when all the threads end).
    ic.immediate_state = "(" + " && ".join(immediate_states) + ")"
    return ic


def codegen_await(
    statement: ast.AwaitStatement, oc: OuterContext
) -> InnerContext:
    """
    Specialized variant of codegen_statement for await statements.
    """
    # The Inner Context we will return.
    ic = InnerContext()
    # We build the code for the inner expression.
    inner_expression = codegen_expression(statement.expression)
    # We need a new state for the await statement.
    await_state = SSAGenerator.new_name("await")
    # We build the expected source code.
    ic.source_code = oc.indent_str + f"{await_state} |= {inner_expression};\n"
    # Our immediate state is given by the "await_state".
    ic.immediate_state = await_state
    # We own the new single state associated with the await statement.
    ic.owned_states.append(await_state)
    return ic


def codegen_emit(
    statement: ast.EmitStatement, oc: OuterContext
) -> InnerContext:
    """
    Specialized variant of codegen_statement for emit statements.
    """
    # The Inner Context we will return.
    ic = InnerContext()
    # We get the C code for the inner expression (the rvalue).
    inner_expression = codegen_expression(statement.expression)
    # We build the expected source code.
    ic.source_code = (
        oc.indent_str + f"*{statement.signal} = {inner_expression};\n"
    )
    # We return the expected triple. An emit statement always exits instantly.
    ic.immediate_state = "1"
    return ic


def codegen_local(
    statement: ast.EmitStatement, oc: OuterContext
) -> InnerContext:
    """
    Specialized variant of codegen_statement for local statements.
    """
    # The Inner Context we will return.
    ic = InnerContext()
    # We add a new local to our InnerContext based on the name provided in the
    # source code.
    ic.owned_locals.append(statement.name)
    # There is no code associated with a local statement.
    # We return the expected triple. A local statement always exits instantly.
    ic.immediate_state = "1"
    return ic


def codegen_if(statement: ast.IfStatement, oc: OuterContext) -> InnerContext:
    """
    Specialized variant of codegen_statement for if-else statements.
    """
    # The Inner Context we will return.
    ic = InnerContext()
    # We start by building the code for our conditional expression.
    inner_expression = codegen_expression(statement.expression)
    # We also build the code for our indented if statement.
    inner_ic = codegen_statement(
        statement.statement, OuterContext(oc.indent + 1)
    )
    # Inheriting the owned and local of our inner expression.
    ic.inherit(inner_ic)
    # We start by filling the code for the if branch before checking if there is
    # an else.
    ic.source_code += f"{oc.indent_str}if ({inner_expression}) {{\n"
    ic.source_code += inner_ic.source_code
    ic.source_code += f"{oc.indent_str}}}\n"
    if statement.else_statement is None:
        # No else statement, this is the easy case. The immediate state of the
        # if statement is that of its if branch if it is taken, otherwise it
        # exits immediately.
        ic.immediate_state = (
            f"(!{inner_expression} || {inner_ic.immediate_state} * "
            f"{inner_expression})"
        )
        # We simply return the InnerContext we have built so far.
        return ic
    else:
        # We must add the code for the else statement.
        else_ic = codegen_statement(
            statement.else_statement, OuterContext(oc.indent + 1)
        )
        # We also inherit the InnerContext from the else branch.
        ic.inherit(else_ic)
        # We add the code for the else branch.
        ic.source_code += (
            f"{oc.indent_str}else {{\n"
            f"{else_ic.source_code}"
            f"{oc.indent_str}}}\n"
        )
        # We return the Inner Context we have built.
        return ic


def codegen_submodule(
    statement: ast.Submodule, oc: OuterContext
) -> InnerContext:
    """
    Specialized variant of codegen_statement for submodule statements.
    """
    # The inner context we will return.
    ic = InnerContext()
    # All the argument signals are already pointers, which is what the callee
    # expects, hence we may directly build the code for the arguments.
    code_arguments = ", ".join([str(signal) for signal in statement.arguments])
    # We then simply call the the C function associated with the module by name.
    ic.source_code += f"{oc.indent_str}{statement.name}({code_arguments});\n"
    # A submodule call is always instantaneous, like an emit.
    ic.immediate_state = "1"
    # We add the new submodule to the owned submodules.
    ic.owned_submodules.append(statement.name)
    # We return the newly built context.
    return ic


##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################



#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts:
# Creation Date: 2021-03-12
# Language: Python3

################################### IMPORTS ####################################

# Standard library
# Your imports from the standard library go here


# External imports
# Your imports from other packages go here


# Internal imports
from compass.ast.statement import Statement  # Used for inheritance
from compass.ast.expression import Expression  # Used for type hints

################################### CLASSES ####################################


class AwaitStatement(Statement):
    """
    Class used to represent an await statement in the AST.
    """

    def __init__(self, expression: Expression):
        """
        Constructor of the Await class.

        Arguments
        =========
         - expression: The Expression to wait for.
        """
        # Storing the arguments.
        self.expression = expression


################################## FUNCTIONS ###################################

# Your functions go here

##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################



#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts:
# Creation Date: 2021-03-12
# Language: Python3

################################### IMPORTS ####################################

# Standard library
from typing import List  # Used for type hints


# External imports
# Your imports from other packages go here


# Internal imports
from compass.ast.node import Node  # Used for inheritance
from compass.ast.declaration import Declaration  # Used for type hints
from compass.ast.statement import Statement  # Used for type hints

################################### CLASSES ####################################


class Module(Node):
    """
    Class used to represent a Module in the AST.
    """

    def __init__(
        self,
        name: str,
        list_declaration: List[Declaration],
        statement: Statement,
    ):
        """
        Constructor of the Module class.

        Arguments
        =========
         - name: The name to use for this Module.
         - list_declaration: The inputs and outputs declared for this Module.
         - statement: The Statement in the body of this Module.
        """
        # Storing the arguments.
        self.name = name
        self.declarations = list_declaration
        self.statement = statement

    def rename(self, name: str):
        """
        Renames the module to the provided value.
        """
        self.name = name


################################## FUNCTIONS ###################################

# Your functions go here

##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################



#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts:
# Creation Date: 2021-03-30
# Language: Python3

################################### IMPORTS ####################################

# Standard library
import json  # Used for json files manipulations
import argparse  # Used for CLI of compass-builder


# External imports
# Your imports from other packages go here


# Internal imports
from compass.frontend.builder_configuration import (
    BuilderConfiguration,
)  # Used for Makefile generation

################################### CLASSES ####################################

# Your classes go here

################################## FUNCTIONS ###################################


def build(json_path: str, makefile_path: str):
    """
    Builds the Makefile to compile the code for the project based on the
    provided JSON specification.

    Arguments
    =========
     - json_path: The path to the JSON containing the description of the
        project.
     - makefile_path: The file to which the Makefile should be written.
    """
    # We first read the JSON file.
    with open(json_path, "r") as json_file:
        json_content = json.load(json_file)
    # We deserialize the content of the JSON file.
    builder_configuration = BuilderConfiguration.from_json(json_content)
    # We build the code for the Makefile.
    makefile_source = builder_configuration.to_makefile()
    # We write the content of the Makefile to the provided path.
    with open(makefile_path, "w") as makefile_file:
        makefile_file.write(makefile_source)


def command_line():
    """
    Entry point for the CLI of compass-builder.
    """
    # We build the argparse parser in this method because it is quite short.
    # We add an epilog about examples of the JSON files.
    cli_parser = argparse.ArgumentParser(
        prog="compass-builder",
        epilog="See the documentation for JSON examples.",
    )
    # Adding the arguments compass-builder.
    cli_parser.add_argument(
        "JSON", type=str, help="The JSON description of the project."
    )
    cli_parser.add_argument(
        "MAKEFILE",
        type=str,
        help="The path to which the Makefile should be written.",
    )
    # We parse the provided arguments.
    arguments = cli_parser.parse_args()
    # We call the build function with the right arguments.
    build(arguments.JSON, arguments.MAKEFILE)


##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################



#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts:
# Creation Date: 2021-03-12
# Language: Python3

################################### IMPORTS ####################################

# Standard library
# Your imports from the standard library go here


# External imports
# Your imports from other packages go here


# Internal imports
import compass.ast as ast  # Used for downcasts
from compass.codegen.compass.expression import (
    codegen_expression,
)  # Used for recursive codegen

################################### CLASSES ####################################

# Your classes go here

################################## FUNCTIONS ###################################


def codegen_statement(statement: ast.Statement, indent: int) -> str:
    """
    Outputs the Compass code corresponding to the provided Statement.
    """
    # We quickly compute the indentation string we will need for this statement.
    indent_str = "\t" * indent
    # Several cases appear, one for each kind of statements.
    if isinstance(statement, ast.Each):
        # We build the code for the inner body of the for-each loop.
        inner_body = codegen_statement(statement.statement, indent + 1)
        # We build the code for the reset expression.
        inner_expression = codegen_expression(statement.expression)
        # We return the code for the whole for-each loop.
        return f"{indent_str}each {inner_expression}\n{inner_body}"
    elif isinstance(statement, ast.Seq):
        # We build the code for each inner statement.
        inner_statement_bodies = [
            codegen_statement(inner_statement, indent + 1)
            for inner_statement in statement.statements
        ]
        # We join the different lines.
        inner_body = ";\n".join(inner_statement_bodies) + ";\n"
        # We return the code for the Seq statement.
        return indent_str + f"seq {{\n{inner_body}" + indent_str + "}"
    elif isinstance(statement, ast.Par):
        # We build the code for each inner statement.
        inner_statement_bodies = [
            codegen_statement(inner_statement, indent + 1)
            for inner_statement in statement.statements
        ]
        # We join the different lines.
        inner_body = ";\n".join(inner_statement_bodies) + ";\n"
        # We return the code for the Par statement.
        return indent_str + f"par {{\n{inner_body}" + indent_str + "}"
    elif isinstance(statement, ast.IfStatement):
        # The condition to branch on.
        inner_expression = codegen_expression(statement.expression)
        # The inner if branch.
        inner_if = codegen_statement(statement.statement, indent + 1)
        # Two cases appear, depending whether an else statement was also used.
        if statement.else_statement is None:
            return (
                f"{indent_str}if ({inner_expression})\n{inner_if}\n"
                f"{indent_str}endif"
            )
        else:
            # The code for the else branch.
            inner_else = codegen_statement(statement.else_statement, indent + 1)
            return (
                f"{indent_str}if ({inner_expression})\n{inner_if}\n"
                f"{indent_str}else\n{inner_else}\n"
                f"{indent_str}endif"
            )
    elif isinstance(statement, ast.Submodule):
        # We create code for the list of arguments.
        code_arguments = ", ".join(
            [str(signal) for signal in statement.arguments]
        )
        # We return the code that calls the submodule with the right arguments.
        return f"{indent_str}submodule {statement.name}({code_arguments})"
    elif isinstance(statement, ast.AwaitStatement):
        # Getting the expression that we have to await.
        inner_expression = codegen_expression(statement.expression)
        # We return the code for the await statement.
        return indent_str + f"await {inner_expression}"
    elif isinstance(statement, ast.EmitStatement):
        # We build the code for the inner expression.
        inner_expression = codegen_expression(statement.expression)
        # We return the code for the emit statement.
        return indent_str + f"emit {statement.signal} <- {inner_expression}"
    elif isinstance(statement, ast.LocalStatement):
        # We return the code for the local variable declaration.
        return indent_str + f"local {statement.name}"
    else:
        raise ValueError(f"Unknown statement {statement}")


##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################



#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts:
# Creation Date: 2021-03-12
# Language: Python3

################################### IMPORTS ####################################

# Standard library
# Your imports from the standard library go here


# External imports
# Your imports from other packages go here


# Internal imports
from compass.ast.await_statement import AwaitStatement
from compass.ast.each import Each
from compass.ast.emit_statement import EmitStatement
from compass.ast.expression import Expression
from compass.ast.extern import Extern
from compass.ast.if_statement import IfStatement
from compass.ast.input_declaration import InputDeclaration
from compass.ast.local_statement import LocalStatement
from compass.ast.module import Module
from compass.ast.node import Node
from compass.ast.number import Number
from compass.ast.output_declaration import OutputDeclaration
from compass.ast.par import Par
from compass.ast.seq import Seq
from compass.ast.signal import Signal
from compass.ast.signal_expression import SignalExpression
from compass.ast.declaration import Declaration
from compass.ast.statement import Statement
from compass.ast.submodule import Submodule
from compass.ast.tree import Tree
from compass.ast.unary import *
from compass.ast.binary import *

################################### CLASSES ####################################

# Your classes go here

################################## FUNCTIONS ###################################

# Your functions go here

##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################



#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts:
# Creation Date: 2021-03-13
# Language: Python3

################################### IMPORTS ####################################

# Standard library
import logging  # Used to log some information for the user.


# External imports
# Your imports from other packages go here


# Internal imports
import compass.ast as ast  # Used for downcasts

################################### CLASSES ####################################

# Your classes go here

################################## FUNCTIONS ###################################

# Source code for a C function that reads the next word from stdin, used in
# codegen_tree.
next_word_C_code = """
int next_word(char *buffer, size_t buffer_size) {
    size_t cursor = 0;
    int next_char;
    int working = 1;
    while (working && ((next_char = getchar()) != EOF)) {
        switch (next_char) {
            case '\\n':
            case ' ':
                buffer[cursor] = '\\0';
                working = 0;
                break;
            default:
                if (cursor < buffer_size) {
                    buffer[cursor] = next_char;
                    cursor++;
                }
                else {
                    buffer[buffer_size-1] = '\\0';
                    working = 0;
                }
                break;
        }
    }
    return (next_char != EOF);
}
"""


def codegen_tree(tree: ast.Tree) -> str:
    """
    Outputs the C source code for debugging the provided AST through the command
    line.
    """
    source_code = ""
    # Adding common inclusions for the code.
    source_code += (
        "#include <stdio.h>\n" "#include <stdlib.h>\n" "#include <string.h>\n"
    )
    # Adding an include for the Compass module, assuming the default header name
    # has been used.
    assumed_header = f"{tree.module.name}_compass.h"
    logging.info(f"Assuming header used is {assumed_header}")
    source_code += f'#include "{assumed_header}"\n'

    # Newline for the style.
    source_code += "\n"

    # Forward declaration of the next_word function, used to read the input word
    # by word.
    source_code += "int next_word(char *buffer, size_t buffer_size);\n"

    # Newline for the style.
    source_code += "\n"

    # Starting the main function.
    source_code += "int main(int argc, const char **argv) {\n"
    # We grab the signals from the module.
    module_signals = [
        declaration.signal.name for declaration in tree.module.declarations
    ]
    # We also make separate lists for the input and output signals.
    module_inputs = [
        declaration.signal.name
        for declaration in tree.module.declarations
        if isinstance(declaration, ast.InputDeclaration)
    ]
    module_outputs = [
        declaration.signal.name
        for declaration in tree.module.declarations
        if isinstance(declaration, ast.OutputDeclaration)
    ]
    # Allocating the signals on the stack.
    source_code += "\tint " + ", ".join(module_signals) + ";\n"
    # To decide what buffer size to use, we pick the longest signal name + 1.
    buffer_size = max(len(signal) for signal in module_signals) + 1
    # We allocate some memory on the stack for the names (which are assumed to
    # be rather short).
    source_code += f"\tchar word[{buffer_size}];\n"
    # Newline for the style.
    source_code += "\n"
    # We enter the event loop, which starts by resetting all the signals.
    source_code += "\twhile(1) {\n"
    for signal in module_signals:
        source_code += f"\t\t{signal} = 0;\n"

    # We read the next word from the user.
    source_code += (
        "\n"  # For the style
        f"\t\tif (!next_word(word, {buffer_size})) {{\n"
        f"\t\t\tword[{buffer_size - 1}] = '\\0';\n"
        '\t\t\tfprintf(stderr, "Encountered EOF after %s\\n", word);\n'
        "\t\t\treturn -1;\n"
        "\t\t}\n"
        "\n"  # For the style
    )

    # Setting the right signal depending on the user input.
    for index, signal in enumerate(module_inputs):
        # Using if only for the first test.
        if index == 0:
            keyword = "if"
        else:
            keyword = "else if"
        # Adding the source code.
        source_code += (
            f'\t\t{keyword} (strncmp(word, "{signal}", {len(signal)+1}) == '
            "0) {\n"
            f"\t\t\t{signal} = 1;\n"
            "\t\t}\n"
        )
    # Adding the else case.
    source_code += (
        "\t\telse {\n"
        '\t\t\tfprintf(stderr, "Unknown Signal %s\\n", word);\n'
        "\t\t}\n"
        "\n"  # For the style
    )

    # Calling the module.
    source_code += (
        f"\t\t{tree.module.name}("
        + ", ".join([f"&{signal}" for signal in module_signals])
        + ");\n"
        "\n"  # For the style
    )

    # Reporting the values of all the signals.
    source_code += '\t\tputs("Signal Report\\n=============");\n'
    for signal in module_signals:
        source_code += f'\t\tprintf("Signal {signal}: %d\\n", {signal});\n'

    # Ending the interaction loop and the main function.
    source_code += (
        "\t}\n" "\treturn EXIT_SUCCESS;\n" "}\n" "\n"  # Newline for the style
    )

    # Adding the source code for next_word.
    source_code += next_word_C_code

    # Returning the built source code.
    return source_code


##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################



#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts:
# Creation Date: 2021-03-12
# Language: Python3

################################### IMPORTS ####################################

# Standard library
# Your imports from the standard library go here


# External imports
# Your imports from other packages go here


# Internal imports
from compass.frontend.driver import compass_compile
from compass.frontend.command_line import command_line
# We give a more explicit name to the compass-builder entry point.
from compass.frontend.builder import command_line as compass_builder

################################### CLASSES ####################################

# Your classes go here

################################## FUNCTIONS ###################################

# Your functions go here

##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################



#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts:
# Creation Date: 2021-03-16
# Language: Python3

################################### IMPORTS ####################################

# Standard library
# Your imports from the standard library go here


# External imports
# Your imports from other packages go here


# Internal imports
from compass.ast.statement import Statement  # Used for inheritance

################################### CLASSES ####################################


class LocalStatement(Statement):
    """
    The "local" statement used to create local variables in the code.
    """

    def __init__(self, name: str):
        """
        Constructor of the LocalStatement class.

        Arguments
        =========
         - name: The name to use for the local variable.
        """
        # Storing the arguments
        self.name = name


################################## FUNCTIONS ###################################

# Your functions go here

##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################



#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts:
# Creation Date: 2021-03-12
# Language: Python3

################################### IMPORTS ####################################

# Standard library
# Your imports from the standard library go here


# External imports
# Your imports from other packages go here


# Internal imports
from compass.ast.declaration import Declaration  # Used for inheritance.
from compass.ast.signal import Signal  # Used for type hints.

################################### CLASSES ####################################


class InputDeclaration(Declaration):
    """
    Class used to represent the declaration of an input Signal in the AST.
    """

    def __init__(self, signal: Signal):
        """
        Constructor of the InputDeclaration class.

        Arguments
        =========
         - signal: The Signal which will be treated as an input to the module.
        """
        # Storing our argument.
        self.signal = signal


################################## FUNCTIONS ###################################

# Your functions go here

##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################



#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts:
# Creation Date: 2021-03-14
# Language: Python3

################################### IMPORTS ####################################

# Standard library
# Your imports from the standard library go here

# External imports
# Your imports from other packages go here

# Internal imports
from compass.ast.expression import (
    Expression,
)  # Used for inheritance and type hints.

################################### CLASSES ####################################


class Unary(Expression):
    """
    Base class used to represent unary operators in the AST.
    """

    def __init__(self, expression: Expression):
        """
        Constructor of the base Unary class.

        Arguments
        =========
         - expression: The expression to apply the unary operator to.
        """
        # Storing the arguments.
        self.expression = expression


################################## FUNCTIONS ###################################

# Your functions go here

##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################



#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts:
# Creation Date: 2021-03-12
# Language: Python3

################################### IMPORTS ####################################

# Standard library
# Your imports from the standard library go here


# External imports
# Your imports from other packages go here


# Internal imports
import compass.ast as ast  # Used for downcasts
from compass.codegen.compass.declaration import (
    codegen_declaration,
)  # Used for recursive codegen
from compass.codegen.compass.statement import (
    codegen_statement,
)  # Used for recursive codegen

################################### CLASSES ####################################

# Your classes go here

################################## FUNCTIONS ###################################


def codegen_module(module: ast.Module) -> str:
    """
    Outputs the Compass code corresponding to the provided Module.
    """
    source_code = ""
    # Adding the name of the module.
    source_code += f"module {module.name}("
    # Building the code for each declaration.
    code_declarations = [
        codegen_declaration(declaration) for declaration in module.declarations
    ]
    # Joining the codes for the declarations.
    source_code += ", ".join(code_declarations) + ')\n'
    # Adding the (indented) code for the body of the module.
    source_code += codegen_statement(module.statement, indent=1) + "\n"
    # Adding the endmodule keyword.
    source_code += "endmodule\n"
    # Returing the prepared source code.
    return source_code


##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################



#!/usr/bin/env python3

################################### METADATA ###################################

# Contributors: roadelou
# Contacts: 
# Creation Date: 2021-03-21
# Language: Python3

################################### IMPORTS ####################################

# Standard library 
from typing import List # Used for type hints


# External imports 
# Your imports from other packages go here 


# Internal imports 
from compass.ast.statement import Statement  # Used for inheritance
from compass.ast.signal import Signal  # Used for type hints

################################### CLASSES ####################################

class Submodule(Statement):
    """
    Class used to represent a submodule call in the AST.
    """

    def __init__(self, name: str, arguments: List[Signal]):
        """
        Constructor of the Submodule class.

        Arguments
        =========
         - name: The name of the submodule that should be called.
         - arguments: The list of Signals to pass to the submodule.
        """
        # Storing the arguments.
        self.name = name
        self.arguments = arguments

################################## FUNCTIONS ###################################

# Your functions go here 

##################################### MAIN #####################################

if __name__ == "__main__":
    # The code to run when this file is used as a script goes here
    pass

##################################### EOF ######################################
